



from kabaret.gui.widgets.views import AbstractGuiView, QtGui, QtCore, resources
from kabaret.core.apps.command import CommandError
from kabaret.core.events.event import Event

FLOW_APP_KEY = 'FLOW'

class NavItem(object):
    SEP = '/'
    
    @classmethod
    def create_root(cls, parent, layout, root_node_id, get_relations, get_cases, on_set_path):
        nav = cls(parent, layout)
        nav._get_relations = get_relations
        nav._get_cases = get_cases
        nav._on_set_path = on_set_path
        nav._override_node_id = root_node_id
        nav.set_choices(get_relations(root_node_id))
        nav.update_display()
        return nav
    
    def __init__(self, parent, layout, prev=None, value=None):
        super(NavItem, self).__init__()
        self._dead = False

        self._override_node_id = None
        
        self.parent = parent
        self.layout = layout
        
        self.prev = prev
        if self.prev is not None:
            self.prev.next = self
        self.next = None
        
        self.w = self._build_widget()
        
        self.value = value
        self.choices = []
        
        if prev:
            self._get_relations = prev._get_relations
            self._get_cases = prev._get_cases
            self._on_set_path = prev._on_set_path
            self.set_choices(self._fetch_choices())

        self.update_display()

        if not self.is_sep():
            self.append(None)

    def _build_widget(self):
        pb = QtGui.QPushButton(self.parent)
        
        self.layout.addWidget(pb)
        
        menu = QtGui.QMenu(pb)
        pb.setMenu(menu)

        pb.setStyleSheet(
        '''        
        QPushButton {
            padding: 2px;
            margin: 0px;
            height: 1em;
            border: 0px;
            /*border: 3px solid red;*/
            border-radius: 0px;
            background: transparent;
        }
        QPushButton[is_leaf_sep="true"] {
            border: 1px outset palette(midlight);
        }
        QPushButton[is_leaf_sep="false"] {
            border: none;
        }
        QPushButton:hover {
            padding: 1px;
            border: 1px inset palette(midlight);
        }
        QPushButton:menu-indicator {
            image: none;
            width: 0px;
        }        
        '''
        )

        return pb

    def _fetch_choices(self):
        return self._get_relations(self.node_id())
    
    def set_choices(self, choices):
        self.choices = choices
        first_choice = choices and choices[0] or None
        nb_choices = len(choices)
        if self.value not in self.choices:
            self.choices.insert(0, self.value)

        menu = self.w.menu()
        menu.clear()
        for value in self.choices:
            menu.addAction(value and value[1] or self.SEP, lambda value=value: self._on_select(value))
        
        self.do_auto_next()
    
    def do_auto_next(self):
        pass
            
    def update_display(self):
        if self._dead:
            return

        self.w.setToolTip('%s %r %r'%(self.__class__.__name__, self.value, self.path()))
        self.w.setProperty('is_leaf_sep', False)
        if self.is_sep():
            if self.is_leaf():
                self.w.setProperty('is_leaf_sep', True)
            self.w.setText(self.SEP)
        else:
            self.w.setText(self.value[1])
        self.w.setStyleSheet(self.w.styleSheet())
    
    def get_next_class(self, value):
        return self.__class__

    def append(self, value):
        if self.next:
            self.next.drop()

        nav_class = self.get_next_class(value)        
        new_nav = nav_class(
            self.parent, self.layout, prev=self, value=value
        )
    
    def drop(self):
        self._dead = True
        if self.next:
            self.next.drop()
        if self.prev:
            self.prev.next = None
        self.w.deleteLater()

    def set_leaf(self):
        if self.next:
            self.next.drop()
        self.update_display()

    def _on_select(self, value):
        self.set_value(value)
        
    def set_value(self, value):
        if value == self.value:
            return

        if value is None:
            self.drop()
        
        self.value = value
        self.set_leaf()
        if self.choices:
            self.append(None)
            
        self.update_display()
        
        if self._on_set_path:
            self._on_set_path(self.leaf().path(), self.leaf().node_id())
    
    def node_id(self):
        if self._override_node_id is not None:
            print 300*'#', self._override_node_id
            return self._override_node_id
        
        if self.is_sep():
            return self.prev and self.prev.node_id() or (FLOW_APP_KEY,) 
        return (
            self.prev and self.prev.node_id() or (FLOW_APP_KEY,) 
            + 
            [self.value[1]]
        )

    def path(self):
        if self.is_sep():
            return self.prev and self.prev.path() or [] 
        return (
            self.prev and self.prev.path() or [] 
            + 
            [self.value[1]]
        )

    def is_leaf(self):
        return self.next is None and True or False
    
    def leaf(self):
        if self.next:
            return self.next.leaf()
        return self
    
    def is_root(self):
        return self.prev is None and True or False
    
    def root(self):
        if self.prev:
            return self.prev.root()
        return self
    
    def is_sep(self):
        return self.value is None

class NavNode(NavItem):

    def get_next_class(self, value):
        return NavNodeSep
    
    def node_id(self):
        return self.prev.node_id() + (self.value[1],)

    def path(self):
        return self.prev.path() + [self.value[1]]

    def _fetch_choices(self):
        return self._get_relations(self.prev.node_id())

class NavNodeSep(NavNode):
    def get_next_class(self, value):
        if value is None:
            1/0
        
        value_type = value[0]
        if value_type == 'Cases':
            nav_class = NavCases
        elif value_type == 'ChildNode':
            nav_class = NavNode

        return nav_class
    
    def set_choices(self, choices):
        if not choices:
            self.drop()
        super(NavNodeSep, self).set_choices(choices)
    
    def node_id(self):
        if self._override_node_id is not None:
            return self._override_node_id
        return self.prev and self.prev.node_id() or (FLOW_APP_KEY,)

    def path(self):
        return self.prev and self.prev.path() or []

    def _fetch_choices(self):
        #return super(NavNodeSep, self)._fetch_choices()
        return self._get_relations(self.node_id())

    def set_value(self, value):
        if value is None:
            self.set_leaf()
        else:
            self.append(value)
        self.update_display()
        
        if self._on_set_path:
            self._on_set_path(self.leaf().path(), self.leaf().node_id())


class NavCases(NavItem):

    def get_next_class(self, value):
        return NavCasesSep
    
    def node_id(self):
        return self.prev and self.prev.node_id() or (FLOW_APP_KEY,)

    def path(self):
        return self.prev and self.prev.path() or []

    def _fetch_choices(self):
        return self.prev._fetch_choices()

class NavCasesSep(NavItem):
    SEP = ':'
    
    def get_next_class(self, value):
        return NavCase

    def do_auto_next(self):
        if self.next is None and len(self.choices)==2:
            self.set_value(self.choices[1])

    def node_id(self):
        return self.prev and self.prev.node_id() or (FLOW_APP_KEY,)

    def path(self):
        return self.prev and self.prev.path() or []

    def _fetch_choices(self):
        return self._get_cases(self.node_id(), self.prev.value[1])
    
    def set_value(self, value):
        if value is None:
            self.set_leaf()
        else:
            self.append(value)
        self.update_display()
        
        if self._on_set_path:
            self._on_set_path(self.leaf().path(), self.leaf().node_id())

class NavCase(NavNode):

    def node_id(self):
        return self.prev.node_id() + (':'.join((self.prev.prev.value[1], self.value[1])),)

    def path(self):
        return self.prev.path() + [self.prev.prev.value[1], self.value[1]]

    def _fetch_choices(self):
        return self.prev._fetch_choices()


class NavPath(QtGui.QWidget):
    def __init__(self, parent, get_relations, get_cases, on_set_path):
        super(NavPath, self).__init__(parent)
        
        self.setLayout(QtGui.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0);
        self.layout().setSpacing(0);
        
        self._getters = (get_relations, get_cases, on_set_path)
        self.root_nav = None
    
    def reset(self, root_node_id):
        if self.root_nav is not None:
            self.root_nav.drop()
        self.root_nav = NavNodeSep.create_root(
            self, self.layout(), root_node_id, *self._getters
        )
        
class NavView(AbstractGuiView):
    def __init__(self, parent, client, app_name):
        super(NavView, self).__init__(parent, client, app_name)

        self.setLayout(QtGui.QHBoxLayout())

        self.nav_bar = NavPath(
            self, 
            self.get_relations,
            self.get_cases,
            self.on_nav_set_path
        )
        self.layout().addWidget(self.nav_bar)
        self.layout().addStretch(100)
    
    def on_connect(self):
        root_node_id = self.client.apps.FLOW.cmds.GetRootId()
        self.nav_bar.reset(root_node_id)
        
    def get_relations(self, node_id):
        if not self.client.connected():
            return []
        
        try:
            relations = self.client.apps.FLOW.cmds.GetNodeRelations(node_id=node_id, child_nodes=True)
        except CommandError:
            import traceback
            traceback.print_exc()
            return []
        
        flat_relations = []
        for relation_type_name, relation_names in relations.items():
            if relation_type_name == 'ChildNode':
                for relation_name in relation_names:
                    flat_relations.append(
                        (relation_type_name, relation_name['name'])
                    )
            else:
                for relation_name in relation_names:
                    flat_relations.append((relation_type_name, relation_name))
        
        return flat_relations
    
    def get_cases(self, node_id, relation_name):
        if not self.client.connected():
            return []

        try:
            case_ids = self.client.apps.FLOW.cmds.GetCaseIds(node_id=node_id, relation_name=relation_name)
        except CommandError:
            import traceback
            traceback.print_exc()
            return []
        return [ ('Case', case_id) for case_id in case_ids ]
    
    def on_nav_set_path(self, path, node_id):
        self.client.send_event(
            Event('GUI', ['FLOW', 'Nav', 'current'], Event.TYPE.UPDATED, node_id)
        )

    
    
    


